VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "CCPLFiles"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
  Option Explicit
  ' demo project that shows how to enumerate control panel applets.
  ' this project also shows how to create a thunking layer using C code
  ' and the thunk compiler ("Thunk.exe" found in the Win32 SDK tools) to
  ' allow the enumeration of 16bit applets on Win9x systems.  the thunk
  ' layer is not necessary for applet enumeration on NT systems.
  '
  ' for more info on thunking see the "Thunks.txt" or "Thunks.doc" file
  ' included in this demo package.
  '
  ' this was written by Bryan Stafford of New Vision Software - newvision@mvps.org.
  ' all of the code contained in this project is copyrighted.  see the
  ' "House Rules" setion at http://www.mvps.org/vbvision for info on the
  ' allowed usage of the contained code.
  '
  '               Copyright  2000  Bryan Stafford
  '
  ' I also wish to acknoledge the assistence provided by Brad Martines
  ' (http://www.mvps.org/btmtz) in the design and debugging of portions
  ' of this project.  Thanks Brad!
  '


  ' small and large system imagelist icon sizes
  Private m_cxSmIcon As Long
  Private m_cySmIcon As Long
  Private m_cxLgIcon As Long
  Private m_cyLgIcon As Long

  Private m_udtIID_IDispatch As GUID   ' IDispatch Interface ID

  ' struct to hold info about each applet
  Private Type APPLET_INFO
    sPath As String
    sTitle As String
    sDescription As String
    sFileName As String
    nAppletIndex As Long
    bIs16Bit As Boolean
    lData As Long
    eType As CPLTypes
    bFailedCPL_INIT As Boolean
    picSmallIcon As StdPicture   ' SmallIcon property
    picLargeIcon As StdPicture   ' LargeIcon property
  End Type
  
  ' array to holde info about each of the applets.  the members of each array item
  ' can be accessed via the public properties of this object by passing the desired
  ' array index when accessing the property
  Private m_audtAppletInfo() As APPLET_INFO
  
  ' current index to add to the applet info array
  Private m_nCurrentIndex As Long

  ' variable to hold the version info for the current OS
  Private m_sngOSVersion As OS_VERSIONS

  ' object that returns info about each applet on the system
  Private WithEvents m_oAppletInfo As cGetAppletInfo
Attribute m_oAppletInfo.VB_VarHelpID = -1

Private Sub Class_Initialize()

  Dim himl As Long
  Dim sfi As SHFILEINFO
  Dim pidlDesktop As Integer

  ' cache the OS version
  m_sngOSVersion = GetOSVersion()
  
  ' Get the system imagelist large and small icon sizes. Any valid
  ' file system path or PIDL can be used to retrieve system image
  ' list handles, in this case we're passing the Desktop's PIDL
  ' (which is just a pointer to 2 bytes of zero-d memory)
  himl = SHGetFileInfo(VarPtr(pidlDesktop), 0, sfi, Len(sfi), SHGFI_SYSICONINDEX Or SHGFI_SMALLICON)
  If himl Then
    Call ImageList_GetIconSize(himl, m_cxSmIcon, m_cySmIcon)
  End If
  
  ' if something went wrong, guess (we should use
  ' GetSystemMetrics/SM_CX/YSMICON, but what the heck...)
  If (m_cxSmIcon = 0) Or (m_cySmIcon = 0) Then
    m_cxSmIcon = 16
    m_cySmIcon = 16
  End If

  himl = SHGetFileInfo(VarPtr(pidlDesktop), 0, sfi, Len(sfi), SHGFI_SYSICONINDEX Or SHGFI_LARGEICON)
  If himl Then
    Call ImageList_GetIconSize(himl, m_cxLgIcon, m_cyLgIcon)
  End If
  
  If (m_cxLgIcon = 0) Or (m_cyLgIcon = 0) Then
    m_cxLgIcon = 32
    m_cxLgIcon = 32
  End If
  
  ' Fill the IDispatch Interface ID, {00020400-0000-0000-C000-000000046}
  With m_udtIID_IDispatch
    .Data1 = &H20400
    .Data4(0) = &HC0
    .Data4(7) = &H46
  End With


  ' find all of the available applets on the system....
  GetApplets
  
End Sub

Private Sub Class_Terminate()
  ' make sure everything gets cleaned up....
  
  Erase m_audtAppletInfo
  
  Set m_oAppletInfo = Nothing
End Sub

' note that there is no error handling in the properties that access the members of the
' module level array.  these will raise errors if they are accesses before the array
' is initialized.  adding error handlers is left as an exorcize for the reader.

' the total number of applets
Public Property Get Count() As Long
  Count = UBound(m_audtAppletInfo()) + 1
End Property

' the properties of each array element.  they are accesses by passing in the desired array index
Public Property Get Path(ByVal nIndex&) As String
  Path = m_audtAppletInfo(nIndex).sPath
End Property

Public Property Get Title(ByVal nIndex&) As String
  Title = m_audtAppletInfo(nIndex).sTitle
End Property

Public Property Get Description(ByVal nIndex&) As String
  Description = m_audtAppletInfo(nIndex).sDescription
End Property

Public Property Get FileName(ByVal nIndex&) As String
  FileName = m_audtAppletInfo(nIndex).sFileName
End Property

Public Property Get AppletIndex(ByVal nIndex&) As Long
  AppletIndex = m_audtAppletInfo(nIndex).nAppletIndex
End Property

Public Property Get Is16Bit(ByVal nIndex&) As Boolean
  Is16Bit = m_audtAppletInfo(nIndex).bIs16Bit
End Property

Public Property Get lData(ByVal nIndex&) As Long
  lData = m_audtAppletInfo(nIndex).lData
End Property

Public Property Get AppletType(ByVal nIndex&) As CPLTypes
  AppletType = m_audtAppletInfo(nIndex).eType
End Property

Public Property Get FailedInit(ByVal nIndex&) As Boolean
  FailedInit = m_audtAppletInfo(nIndex).bFailedCPL_INIT
End Property

Public Property Get SmallIcon(ByVal nIndex&) As StdPicture
  Set SmallIcon = m_audtAppletInfo(nIndex).picSmallIcon
End Property

Public Property Get LargeIcon(ByVal nIndex&) As StdPicture
  Set LargeIcon = m_audtAppletInfo(nIndex).picLargeIcon
End Property

Public Sub Refresh()
  GetApplets
End Sub

Private Function GetApplets() As Boolean
  ' enumerate all of the control panel applets on the system....
  
  Dim nTotalItems As Long
  Dim asApplets() As String
  Dim oGetApplets As cFindApplets

  ' initialize the module level array
  ReDim m_audtAppletInfo(20) As APPLET_INFO
  
  ' reset the var that holds the current index to add to the info array
  m_nCurrentIndex = 0
  
  Set oGetApplets = New cFindApplets
  
  ' fill our array with the paths to each control panel applet lib on the system
  oGetApplets.GetAppletPaths asApplets()
  
  Set oGetApplets = Nothing
  
  Set m_oAppletInfo = New cGetAppletInfo
  
  nTotalItems = m_oAppletInfo.EnumerateApplets()
  
  
  If nTotalItems = (-1) Then
    ReDim m_audtAppletInfo(nTotalItems To nTotalItems) As APPLET_INFO
  Else
    ' remove any extra indexes from our info array
    ReDim Preserve m_audtAppletInfo(0 To nTotalItems) As APPLET_INFO
  End If
  
  Set m_oAppletInfo = Nothing
  
End Function

Private Sub m_oAppletInfo_ReturnAppletInfo(ByVal sPath As String, ByVal nIndex As Long, ByVal sInternalName As String, ByVal sDescription As String, _
                        ByVal hIcon As Long, ByVal lData As Long, ByVal eType As CPLTypes, ByVal bFailedInit As Boolean, ByVal bIs16Bit As Boolean, ByVal nIconIndex As Long)
  ' this is an event that is raised in from the module level AppletInfo object.  it allows us to
  ' format the info and add it to our module level array
  
  Dim nUbound As Long
  Dim hLargeIcon As Long
  Dim hSmallIcon As Long
  Dim hIconCopy As Long
  
  nUbound = UBound(m_audtAppletInfo)
  
  If m_nCurrentIndex > nUbound Then ReDim Preserve m_audtAppletInfo(0 To (nUbound + 20)) As APPLET_INFO
  
  With m_audtAppletInfo(m_nCurrentIndex)
    .sPath = sPath
    .sTitle = RemoveMnemonic(sInternalName)
    .sDescription = sDescription
    .sFileName = AppletNameFromPath(sPath)
    .nAppletIndex = nIndex
    .bFailedCPL_INIT = bFailedInit
    .bIs16Bit = bIs16Bit
    .lData = lData
    .eType = eType
    
    If hIcon Then
      ' make a local copy the CPL's icon resource handle (it will be invalid when the hMod is freed)
      hIconCopy = CopyImage(hIcon, IMAGE_ICON, m_cxSmIcon, m_cySmIcon, LR_COPYFROMRESOURCE)
      If hIconCopy Then
        ' have the picture own the icon handle (allows APIs to work with Picture.Handle)
        Set .picSmallIcon = GetPicture(hIconCopy, vbPicTypeIcon, True)
      End If
      hIconCopy = CopyImage(hIcon, IMAGE_ICON, m_cxLgIcon, m_cyLgIcon, LR_COPYFROMRESOURCE)
      If hIconCopy Then
        Set .picLargeIcon = GetPicture(hIconCopy, vbPicTypeIcon, True)
      End If
      
    ElseIf bIs16Bit Then
      If ExtractIconEx(sPath, -nIconIndex, hLargeIcon, hSmallIcon, 1&) Then
        ' make a local copy the CPL's icon resource handle (it will be invalid when the hMod is freed)
        hIconCopy = CopyImage(hSmallIcon, IMAGE_ICON, m_cxSmIcon, m_cySmIcon, LR_COPYFROMRESOURCE)
        If hIconCopy Then
          ' have the picture own the icon handle (allows APIs to work with Picture.Handle)
          Set .picSmallIcon = GetPicture(hIconCopy, vbPicTypeIcon, True)
        End If
        hIconCopy = CopyImage(hLargeIcon, IMAGE_ICON, m_cxLgIcon, m_cyLgIcon, LR_COPYFROMRESOURCE)
        If hIconCopy Then
          Set .picLargeIcon = GetPicture(hIconCopy, vbPicTypeIcon, True)
        End If
      End If
    End If   ' hIcon
  End With
    
  ' increment the current index
  m_nCurrentIndex = m_nCurrentIndex + 1
  
End Sub

Private Function GetPicture(ByVal hImage As Long, ByVal dwPicType As PictureTypeConstants, ByVal fPictureOwnsHandle As Boolean) As StdPicture

  ' Creates a Picture object from an image handle
  '   hImage               - either an HBITMAP, HMETAFILE, HICON or HENHMETAFILE
  '   dwPicType         - one of VB's PictureTypeConstants describing the hImage param.
  '   fPictureOwnsHandle  - flag specifying whether the new Picture owns and destroys hImage
  '                                       when the Picture itself is released. Don't destroy it twice!!! ...or else
  
  ' Returns a VB Picture object on success, or the object value "Nothing" on error.
  
  Dim pd As PICTDESC
  
  If hImage Then
    pd.cbSize = Len(pd)
    pd.PicType = dwPicType
    pd.hImage = hImage
    Call OleCreatePictureIndirect(pd, m_udtIID_IDispatch, Abs(fPictureOwnsHandle), GetPicture)
  End If
  
End Function

' Returns the string before first null char encountered (if any) from an ANSII string.
Private Function GetStrFromBufferA(sz As String) As String
  GetStrFromBufferA = Left$(sz, lstrlenW(ByVal StrPtr(sz)))
End Function

Private Function RemoveMnemonic(ByVal sName As String) As String
  ' removes any "&" chars from the passed string that are there to create underscores in the displayed text
  
  Dim nLoc1 As Long
  Dim nLoc2 As Long
  Dim nLen As Long
  
  nLoc2 = 1
  
  Do
    nLen = Len(sName)
    
    nLoc1 = InStr(nLoc2, sName, "&")
    
    If nLoc1 Then
      If nLoc1 < nLen Then
        If Mid$(sName, nLoc1 + 1, 1) <> " " Then
          sName = Left$(sName, nLoc1 - 1) & Right$(sName, nLen - nLoc1)
          nLoc2 = nLoc1
        End If
      End If
    End If
  Loop Until nLoc1 = 0
  
  RemoveMnemonic = sName
  
End Function

Public Function DisplayApplet(ByVal hWndOwner As Long, ByVal nIndex As Long) As Boolean
  
  Dim bError As Boolean
  Dim hMod As Long
  Dim lpfn As Long
  Const CPlApplet As String = "CPlApplet"    ' lower case "L" in "CPL"
  
  ' dims for 16bit thunking code
  Dim lpMem16 As Long
  Dim hWnd As Long
  Dim nMsg As Long
  Dim lParam1 As Long

  With m_audtAppletInfo(nIndex)
    hMod = LoadLibrary(.sPath)
    If hMod Then
      ' if it loaded, get the address of the CPlApplet function
      lpfn = GetProcAddress(hMod, ByVal CPlApplet)   ' the function name is case sensitive
      If lpfn Then
        ' if we found the address, use the CallWindowProc function to call the CPlApplet function.
        ' first, initialize the lib
        If CallWindowProc(lpfn, hWndOwner, CPL_INIT, 0, 0) = 0 Then
          MsgBox "Applet failed CPL_INIT.  Attempting CPL_DBLCLK message." & vbNewLine & "It may not be possible to display this applet....", vbInformation, "CPlApplet Demo"
        End If
        
        ' tell the applet to display itself as if it were double clicked in the controlpanel
        Call CallWindowProc(lpfn, hWndOwner, CPL_DBLCLK, .nAppletIndex, .lData)

        ' tell this applet to free any resources it may have initialized
        Call CallWindowProc(lpfn, hWndOwner, CPL_STOP, .nAppletIndex, 0)

        ' tell the lib to free all remaining resources
        Call CallWindowProc(lpfn, hWndOwner, CPL_EXIT, 0, 0)
      
      Else
        bError = True
      End If   ' lpfn
      
      ' free the lib
      Call FreeLibrary(hMod)
      
      
    ElseIf m_sngOSVersion < osvWinNT4 Then ' unable to load the lib using the 32bit call above.
      ' this may be a 16 bit lib on Win9x so check the OS version befor trying to load it using the thunked 16bit call
      hMod = LoadLibrary16(.sPath)
      If hMod Then
        lpfn = GetProcAddress16(hMod)
        If lpfn Then
          ' this block isn't needed in this implementation of the code but would be
          ' useful if you were displaying an applet from an app with a user interface.
          If IsWindow(hWndOwner) Then
            hWnd = WOWHandle16(hWnd, WOW_TYPE_HWND)
          Else
            hWnd = 0
          End If
          
          ' first, send the init message
          nMsg = CPL_INIT
          lParam1 = .nAppletIndex
          If CPlApplet16(lpfn, hWnd, nMsg, lParam1, lpMem16) = 0 Then
            MsgBox "Applet failed CPL_INIT.  Attempting CPL_DBLCLK message." & vbNewLine & "It may not be possible to display this applet....", vbInformation, "CPlApplet Demo"
          End If
          
          ' tell the applet to display itself as if it were double clicked in the controlpanel
          nMsg = CPL_DBLCLK
          Call CPlApplet16(lpfn, hWnd, nMsg, lParam1, .lData)
          
          ' tell this applet to free any resources it may have initialized
          nMsg = CPL_STOP
          Call CPlApplet16(lpfn, hWnd, nMsg, lParam1, lpMem16)

          ' tell the module to free all remaining resources
          nMsg = CPL_EXIT
          Call CPlApplet16(lpfn, hWnd, nMsg, lParam1, lpMem16)
     
        Else
          bError = True
        End If   ' lpfn
        ' free the 16bit module
        Call FreeLibrary16(hMod)
      Else
        bError = True
      End If   ' hMod
    Else
      bError = True
    End If   ' hMod - Else
  End With ' m_audtAppletInfo(nIndex)

  If bError Then MsgBox "Unable to display applet.", vbExclamation, "CPLApplet"
  
End Function

Private Function AppletNameFromPath(ByVal sPath As String) As String

  Dim i As Long
  Dim nLastLocation As Long
  Dim nLen As Long
  
  nLen = Len(sPath)
  
  Do
    If nLastLocation < nLen Then
      i = InStr(nLastLocation + 1, sPath, "\", vbTextCompare)
      
      If i Then
        nLastLocation = i
      Else
        Exit Do
      End If
    Else
      Exit Do
    End If
  Loop
  
  If nLastLocation Then AppletNameFromPath = Right$(sPath, nLen - nLastLocation)
    
End Function
