VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "cFindApplets"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
  Option Explicit
  ' demo project that shows how to enumerate control panel applets.
  ' this project also shows how to create a thunking layer using C code
  ' and the thunk compiler ("Thunk.exe" found in the Win32 SDK tools) to
  ' allow the enumeration of 16bit applets on Win9x systems.  the thunk
  ' layer is not necessary for applet enumeration on NT systems.
  '
  ' for more info on thunking see the "Thunks.txt" or "Thunks.doc" file
  ' included in this demo package.
  '
  ' this was written by Bryan Stafford of New Vision Software - newvision@mvps.org.
  ' all of the code contained in this project is copyrighted.  see the
  ' "House Rules" setion at http://www.mvps.org/vbvision for info on the
  ' allowed usage of the contained code.
  '
  '               Copyright  2000  Bryan Stafford
  '
  ' I also wish to acknoledge the assistence provided by Brad Martines
  ' (http://www.mvps.org/btmtz) in the design and debugging of portions
  ' of this project.  Thanks Brad!
  '

  ' variable to hold the version info for the current OS
  Private m_sngOSVersion As OS_VERSIONS

Public Sub GetAppletPaths(asApplets() As String)
  ' this routine looks in all of the places that control panel applets
  ' might be hiding on any given Win32 operating system.  we fill a
  ' module level array with all of the returned paths
  
  Dim i As Long
  Dim j As Long
  Dim nMaxIndex As Long
  Dim nUbound As Long
  Dim nLen As Long
  Dim nCountRemove As Long
  Dim lpSaveAddress As Long
  
  ReDim asApplets(50) As String
  
  nMaxIndex = -1
  
  ' first we check the registry keys that are used in Win2k....
  nMaxIndex = GetCPLKeysFromRegistry(asApplets(), nMaxIndex, HKEY_LOCAL_MACHINE)
  nMaxIndex = GetCPLKeysFromRegistry(asApplets(), nMaxIndex, HKEY_CURRENT_USER)

  ' then we check the system directory for all files with .cpl extensions....
  nMaxIndex = GetSystemDirFileNames(asApplets(), nMaxIndex)
  
  ' next, we need to check the MMCPL section in control.ini....
  nMaxIndex = GetControlIniFielNames(asApplets(), nMaxIndex)
  
  ' after we have all of the file names, we need to redimension
  ' the array to the exact number of entries
  If nMaxIndex < 1 Then nMaxIndex = 1
  ReDim Preserve asApplets(nMaxIndex) As String
  
  ' then we search the array to make sure there are no duplicate items
  For i = 0 To nMaxIndex
    For j = i + 1 To nMaxIndex
      If j > (nMaxIndex - nCountRemove) Then Exit For
      
      nLen = Len(asApplets(i))
      If Len(asApplets(j)) = nLen Then
        If asApplets(j) = asApplets(i) Then
          nCountRemove = nCountRemove + 1
          
          If j >= (nMaxIndex - nCountRemove) Then Exit For
          
          If j < nMaxIndex Then
            ' if this is a duplicate, move it's pointer to the end of the array for later removal....
            lpSaveAddress = StrPtr(asApplets(j))
            CopyMemory ByVal VarPtr(asApplets(j)), ByVal VarPtr(asApplets(j + 1)), (nMaxIndex - j) * 4
            CopyMemory ByVal VarPtr(asApplets(nMaxIndex)), lpSaveAddress, 4&
          End If
          
          ' decrement the counter so that we can check the item that we just moved
          j = j - 1
        End If
      End If
    Next ' j
  Next ' i
  
  ' if there were duplicates, redimension the array to remove them
  If nMaxIndex > nCountRemove Then
    ReDim Preserve asApplets(nMaxIndex - nCountRemove) As String
  End If
  
End Sub
  
Private Function GetCPLKeysFromRegistry(asApplets() As String, ByVal nMaxIndex As Long, Optional ByVal eRootName As REGISTRY_KEYS = HKEY_CURRENT_USER) As Long
  ' returns all values for the key that contains the paths for control panel applets in Win2K and greater systems
  
  Dim lpFT As FILETIME
  Dim hKey As Long
  Dim lRet As Long
  Dim nCurIdx As Long
  Dim sKeyName As String
  Dim sData As String
  Dim nKeyLen As Long
  Dim nDataLen As Long
  Dim nCharLoc1 As Long
  Dim nCharLoc2 As Long
  Dim sEnvironVar As String
  Dim nReserved As Long
  Dim nUbound As Long
  
  Const REG_CPLS_KEY As String = "SOFTWARE\Microsoft\Windows\CurrentVersion\Control Panel\Cpls"
  
  ' check that we can open the key
  If RegOpenKeyEx(eRootName, REG_CPLS_KEY, 0, KEY_QUERY_VALUE, hKey) Then
    GetCPLKeysFromRegistry = nMaxIndex
    Exit Function
  End If

    
  Do ' loop through the values in the key to get all of the registerd applets
    nKeyLen = MAX_PATH
    nDataLen = MAX_PATH
    sKeyName = String$(nKeyLen, 0)
    sData = String$(nDataLen, 0)
    lRet = RegEnumValue(hKey, nCurIdx, sKeyName, nKeyLen, nReserved, 0&, sData, nDataLen)
    

    If lRet = ERROR_SUCCESS Then
      sData = StripNulls(sData)
             
      If Len(sData) Then
        nUbound = UBound(asApplets())
        
        ' the values can contain place holders for special folders such as: "%Program Files%\YourApp\YourCPL.cpl"
        ' we need to resolve these before attempting to load any applets
        nCharLoc1 = InStr(sData, "%")
        If nCharLoc1 Then
          If nCharLoc1 < Len(sData) Then
            nCharLoc2 = InStr(nCharLoc1 + 1, sData, "%")
            If nCharLoc2 Then
              sEnvironVar = Mid$(sData, nCharLoc1 + 1, nCharLoc2 - 2)
              
              ' check to see if the folder is among the environment variables
              sEnvironVar = Environ$(sEnvironVar)
              If Len(sEnvironVar) Then
                sData = sEnvironVar & Right$(sData, Len(sData) - nCharLoc2)
              End If
            End If
          End If
        End If
        
        ' convert the path to 8.3 short format
        sData = Get16bitPath(sData)
        
        If Len(sData) Then
          nMaxIndex = nMaxIndex + 1
        
          If nUbound < nMaxIndex Then ReDim Preserve asApplets(nUbound + 50) As String
          
          ' add the path to the array
          asApplets(nMaxIndex) = sData
        End If
      End If
    End If
      
    nCurIdx = nCurIdx + 1
    
  Loop While lRet = ERROR_SUCCESS
  
  Call RegCloseKey(hKey)

  GetCPLKeysFromRegistry = nMaxIndex
  
End Function

Private Function GetSystemDirFileNames(asApplets() As String, ByVal nMaxIndex As Long) As Long
  ' finds all files with the .cpl extension in the system directory
  
  Dim sSysDir As String
  Dim nChars As String
  Dim sFile As String
  Dim sBuffer As String
  Dim i As Long
  Dim nUbound As Long


  sSysDir = Space$(MAX_PATH)
  nChars = GetSystemDirectory(sSysDir, MAX_PATH)
  If nChars Then
    sSysDir = Left$(sSysDir, nChars)
    If Right$(sSysDir, 1) <> "\" Then sSysDir = sSysDir & "\"
    
    nUbound = UBound(asApplets)
    
    sFile = Dir$(sSysDir & "*.cpl")
    Do While LenB(sFile)
      sBuffer = Get16bitPath(sSysDir & sFile)
      
      If Len(sBuffer) Then
        nMaxIndex = nMaxIndex + 1
      
        If nUbound < nMaxIndex Then
          nUbound = nUbound + 50
          ReDim Preserve asApplets(nUbound) As String
        End If

        asApplets(nMaxIndex) = Get16bitPath(sSysDir & sFile)
      End If
      
      sFile = Dir$
    Loop   ' While LenB(sFile)
  
  End If   ' nChars

  GetSystemDirFileNames = nMaxIndex
  
End Function

Private Function GetControlIniFielNames(asApplets() As String, ByVal nMaxIndex As Long) As Long
  ' returns all entries from the MMCPL section in control.ini
  ' entries are returned as "Key=Path" pairs
  
  Dim nInBuffLen As Long
  Dim nOutBuffLen As Long
  Dim nNextPos As Long
  Dim nUbound As Long
  Dim sBuffer As String
  Dim sPath As String

  nInBuffLen = 1024
  Do
    sBuffer = String$(nInBuffLen, 0)
    
    nOutBuffLen = GetPrivateProfileSection("MMCPL", sBuffer, nInBuffLen, "control.ini")
    If nOutBuffLen = nInBuffLen - 2 Then
      nInBuffLen = nInBuffLen + 1024
      sBuffer = String$(nInBuffLen, 0)
    Else
      Exit Do
    End If
  Loop
  
  If nOutBuffLen Then
    nUbound = UBound(asApplets)
    nNextPos = 1
    
    Do
      sPath = GetPathFromBuffer(sBuffer, nNextPos)
      
      If Len(sPath) Then
        nMaxIndex = nMaxIndex + 1
        
        If nUbound < nMaxIndex Then
          nUbound = nUbound + 50
          ReDim Preserve asApplets(nUbound) As String
        End If
      
        asApplets(nMaxIndex) = sPath
      End If
    Loop Until nNextPos = (-1)
  End If

  GetControlIniFielNames = nMaxIndex
  
End Function

Private Function GetPathFromBuffer(ByRef sBuffer As String, ByRef nStartPos As Long) As String
  ' parses the paths from the passed buffer containing null seperated "Key=Path" pairs
  ' the buffer is the entire "MMCPL" section from control.ini
  
  Dim nNextPos As Long
  Dim nValPos As Long
  Dim sKeyValPair As String
  
  If nStartPos < Len(sBuffer) Then
    nNextPos = InStr(nStartPos, sBuffer, vbNullChar)
    
    If nNextPos > nStartPos + 1 Then
      sKeyValPair = Mid$(sBuffer, nStartPos, nNextPos - nStartPos)
      nStartPos = nNextPos + 1
      
      nValPos = InStr(sKeyValPair, "=")
      If nValPos Then
        GetPathFromBuffer = Get16bitPath(Mid$(sKeyValPair, nValPos + 1, Len(sKeyValPair) - nValPos))
      End If
    Else
      nStartPos = (-1)
    End If
  Else
    nStartPos = (-1)
  End If
  
End Function

Private Function Get16bitPath(ByVal sLongPath As String) As String
  ' returns an 8.3 path string for the passed long patrh if the file exists
  ' only change the path if we are on a Win9x system
  
  Dim sShortPath As String
  Dim nLen As Long
  
  If m_sngOSVersion < osvWinNT4 Then
    Do
      sShortPath = String$(MAX_PATH + 1, 0)
      
      nLen = GetShortPathName(sLongPath, sShortPath, MAX_PATH)
      If (nLen > 0) And (nLen <= (MAX_PATH + 1)) Then
        sShortPath = StripNulls(sShortPath)
      
        Get16bitPath = sShortPath
        
        Exit Do
      End If
    Loop Until nLen = 0
    
  Else
    Get16bitPath = sLongPath
  End If
  
End Function

Private Sub Class_Initialize()
  m_sngOSVersion = GetOSVersion()
End Sub
