VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "cGetAppletInfo"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
  Option Explicit
  ' demo project that shows how to enumerate control panel applets.
  ' this project also shows how to create a thunking layer using C code
  ' and the thunk compiler ("Thunk.exe" found in the Win32 SDK tools) to
  ' allow the enumeration of 16bit applets on Win9x systems.  the thunk
  ' layer is not necessary for applet enumeration on NT systems.
  '
  ' for more info on thunking see the "Thunks.txt" or "Thunks.doc" file
  ' included in this demo package.
  '
  ' this was written by Bryan Stafford of New Vision Software - newvision@mvps.org.
  ' all of the code contained in this project is copyrighted.  see the
  ' "House Rules" setion at http://www.mvps.org/vbvision for info on the
  ' allowed usage of the contained code.
  '
  '               Copyright  2000  Bryan Stafford
  '
  ' I also wish to acknoledge the assistence provided by Brad Martines
  ' (http://www.mvps.org/btmtz) in the design and debugging of portions
  ' of this project.  Thanks Brad!
  '

  
  Public Enum CPLTypes
    cplInvalidApplet = 0&
    cplNewInquireUnicode
    cplNewInquireANSI
    cplOldInquireANSI
  End Enum

  ' variable to hold the version info for the current OS
  Private m_sngOSVersion As OS_VERSIONS

  Event ReturnAppletInfo(ByVal sPath$, ByVal nIndex&, ByVal sInternalName$, _
                  ByVal sDescription$, ByVal hIcon&, ByVal lData&, ByVal eType As CPLTypes, ByVal bFailedInit As Boolean, ByVal bIs16Bit As Boolean, ByVal nIconIndex&)
  
Public Function EnumerateApplets() As Long
  ' enumerate all of the control panel applets on the system
  
  Dim bError As Boolean
  Dim nTotalAdded As Long
  Dim nIndex As Long
  Dim hMod As Long
  Dim lpfn As Long
  Dim bFailedInit As Boolean
  Dim nNumApplets As Long
  Dim i As Long
  Dim ncia As NEWCPLINFOA
  Dim nciw As NEWCPLINFOW
  Dim ci As CPLINFO
  Dim eType As CPLTypes
  Const CPlApplet As String = "CPlApplet"    ' lower case "L" in "CPL"
  
  ' dims for 16bit thunking code
  Dim bIs16Bit As Boolean
  Dim ncia16 As NEWCPLINFO16
  Dim ci16 As CPLINFO16
  Dim lpMem16 As Long
  Dim hMem16 As Long
  Dim lpMem32 As Long
  Dim hWnd As Long
  Dim nMsg As Long
  Dim lParam1 As Long
  Dim idResource As Long
  Dim nLen As Long

  Dim asApplets() As String
  Dim oGetApplets As cFindApplets
  
  Set oGetApplets = New cFindApplets
  
  ' fill our array with the paths to each control panel applet lib on the system
  oGetApplets.GetAppletPaths asApplets()
  
  Set oGetApplets = Nothing
  
  nTotalAdded = (-1)
  
  ' loop through the array and retreive info about each applet
  For nIndex = 0 To UBound(asApplets)
    ' reset the error var for this iteration of the loop
    bError = False
    bFailedInit = False
    bIs16Bit = False
    
    ' load the lib
    hMod = LoadLibrary(asApplets(nIndex))
    If hMod Then
      ' if it loaded, get the address of the CPlApplet function
      lpfn = GetProcAddress(hMod, ByVal CPlApplet)   ' the function name is case sensitive
      If lpfn Then
        ' if we found the address, use the CallWindowProc function to call the CPlApplet function.
        ' first, initialize the lib.  some applets fail this but will still give info when queried
        ' using CPL_GETCOUNT.  we test the init message and add the flag when it fails.
        If CallWindowProc(lpfn, 0, CPL_INIT, 0, 0) = 0 Then bFailedInit = True

        ' then get the number of applets contained in the lib and loop through them
        nNumApplets = CallWindowProc(lpfn, 0, CPL_GETCOUNT, 0, 0)
        
        ' if the init call and the getcount call failed, we need to set the error flag
        If (bFailedInit = True) And (nNumApplets = 0) Then bError = True
        
        For i = 0 To nNumApplets - 1
          ' attempt to fill the wide character (unicode) version of the NEWINQUIRE struct
          nciw.dwSize = 0
          Call CallWindowProc(lpfn, 0, CPL_NEWINQUIRE, i, VarPtr(nciw))
          
          If (nciw.dwSize = Len(nciw)) Then
            ' if the size is correct, all we need to do is format the data and add it to the listbox
            With nciw
              nTotalAdded = nTotalAdded + 1
              
              RaiseEvent ReturnAppletInfo(asApplets(nIndex), i, CStr(.szName), CStr(.szInfo), .hIcon, .lData, cplNewInquireUnicode, bFailedInit, False, 0)
            End With
            
          Else
            ' if the applet didn't fill the wide character version of the struct corectly,
            ' attempt to fill the ANSI version of the NEWINQUIRE struct
            ncia.dwSize = 0
            Call CallWindowProc(lpfn, 0, CPL_NEWINQUIRE, i, VarPtr(ncia))
            
            If (ncia.dwSize <> Len(ncia)) Then
              ' if the NEWINQUIRE call failed, fall back on the INQUIRE call and get the info in the INQUIRE struct
              eType = cplOldInquireANSI
              
              Call CallWindowProc(lpfn, 0, CPL_INQUIRE, i, VarPtr(ci))
              Call LoadString(hMod, ci.idName, ncia.szName(0), 32)
              Call LoadString(hMod, ci.idInfo, ncia.szInfo(0), 64)
              
              ncia.hIcon = LoadImage(hMod, ci.idIcon, IMAGE_ICON, 0, 0, LR_SHARED)
              
              ' place the lData value in the NEWINQUIRE struct
              ncia.lData = ci.lData
            End If
            
            If (eType = cplInvalidApplet) Then eType = cplNewInquireANSI
            ' format the data and add it to the listbox
            With ncia
              nTotalAdded = nTotalAdded + 1
              
              RaiseEvent ReturnAppletInfo(asApplets(nIndex), i, StrConv(.szName, vbUnicode), StrConv(.szInfo, vbUnicode), .hIcon, .lData, eType, bFailedInit, False, 0)
            End With

          End If   ' (nciw.dwSize = Len(nciw))

          ' tell this applet to free any resources it may have initialized
          Call CallWindowProc(lpfn, 0, CPL_STOP, i, 0)

        Next   ' i = 0 To CallWindowProc/CPL_GETCOUNT - 1

        ' tell the lib to free all remaining resources
        Call CallWindowProc(lpfn, 0, CPL_EXIT, 0, 0)
        
      Else
        bError = True
      End If   ' lpfn
      
      ' free the lib
      Call FreeLibrary(hMod)
      
      
    ElseIf m_sngOSVersion < osvWinNT4 Then ' unable to load the lib using the 32bit call above.
      ' this may be a 16 bit lib on Win9x so check the OS version befor trying to load it using the thunked 16bit call
      hMod = LoadLibrary16(asApplets(nIndex))
      If hMod Then
        lpfn = GetProcAddress16(hMod)
        If lpfn Then
          ' allocate and lock some memory in the 16bit heap.
          ' this allows us to pass back the structs from 16bit land
          lpMem16 = WOWGlobalAllocLock16(GPTR, 300&, VarPtr(hMem16))
          If lpMem16 Then
            ' the memory allocation succeeded so now, we need to translate the returned 16bit
            ' pointer to a 32bit pointer.  this will allow us to access the memory from 32bit land.
            lpMem32 = WOWGetVDMPointerFix(lpMem16, 300&, 1&)
            If lpMem32 Then
              ' we have a memory block so now we can make the thunk calls....
              
              ' this block isn't needed in this implementation of the code but would be
              ' useful if you were displaying an applet from an app with a user interface.
              If IsWindow(hWnd) Then
                hWnd = WOWHandle16(hWnd, WOW_TYPE_HWND)
              Else
                hWnd = 0
              End If
              
              ' the 16bit module loaded so this is a 16bit lib
              bIs16Bit = True
              
              ' first, send the init message.  some applets fail this but will still give info when queried
              ' using CPL_GETCOUNT.  we test the init message and add the flag when it fails.
              nMsg = CPL_INIT
              lParam1 = 0
              If CPlApplet16(lpfn, hWnd, nMsg, lParam1, lpMem16) Then bFailedInit = True
  
              ' next, get the number of applets in this module and loop through them
              nMsg = CPL_GETCOUNT
              nNumApplets = CPlApplet16(lpfn, hWnd, nMsg, lParam1, lpMem16)

              ' if the init call and the getcount call failed, we need to set the error flag
              If (bFailedInit = True) And (nNumApplets = 0) Then bError = True

              For i = 0 To nNumApplets - 1
                ' clear out the memory so we aren't fooled by retained values on a call failure
                ZeroMemory ByVal lpMem32, 300&

                ' try the newinquire message first....
                nMsg = CPL_NEWINQUIRE
                lParam1 = i
                Call CPlApplet16(lpfn, hWnd, nMsg, lParam1, lpMem16)
                
                ' copy the data from the 16bit memory block to our local struct
                CopyMemory ByVal VarPtr(ncia16), ByVal lpMem32, Len(ncia)
                
                ' if a newinquire struct wasn't returned, try a regular inquire message
                If (ncia16.dwSize <> Len(ncia16)) Then
                  eType = cplOldInquireANSI
                  
                  nMsg = CPL_INQUIRE
                  Call CPlApplet16(lpfn, hWnd, nMsg, lParam1, lpMem16)
                  
                  ' copy the data from 16bit the memory block to our local struct
                  CopyMemory ByVal VarPtr(ci16), ByVal lpMem32, Len(ci16)

                  ' try to load the resource strings from the 16bit module....
                  lParam1 = 32 '<- the max length of the buffer we want returned. corresponds to the length of ncia16.szName
                  nLen = LoadString16(hMod, ci16.idName, lpMem16, lParam1)
                  ' nLen is the length of the string loaded by the LoadString function.
                  ' copy the string to our local struct.
                  CopyMemory ByVal VarPtr(ncia16.szName(0)), ByVal lpMem32, nLen
                  
                  lParam1 = 64 '<- the max length of the buffer we want returned. corresponds to the length of ncia16.szInfo
                  nLen = LoadString16(hMod, ci16.idInfo, lpMem16, lParam1)
                  CopyMemory ByVal VarPtr(ncia16.szInfo(0)), ByVal lpMem32, nLen
                  
                  ' place the lData value in the NEWINQUIRE struct
                  ncia16.lData = ci16.lData
                End If
                
                If (eType = cplInvalidApplet) Then eType = cplNewInquireANSI
                
                ' format the data and add it to the listbox
                With ncia16
                  nTotalAdded = nTotalAdded + 1
                  
                  RaiseEvent ReturnAppletInfo(asApplets(nIndex), i, StrConv(.szName, vbUnicode), StrConv(.szInfo, vbUnicode), .hIcon, .lData, eType, bFailedInit, True, ci16.idIcon)
                End With
                
                ' tell this applet to free any resources it may have initialized
                nMsg = CPL_STOP
                Call CPlApplet16(lpfn, hWnd, nMsg, lParam1, lpMem16)
                
              Next   ' i = 0 To CPlApplet16/CPL_GETCOUNT - 1

              ' tell the module to free all remaining resources
              nMsg = CPL_EXIT
              Call CPlApplet16(lpfn, hWnd, nMsg, lParam1, lpMem16)
              
              ' un-fix the 32bit pointer
              WOWGetVDMPointerUnfix lpMem16
            Else
              bError = True
            End If ' lpMem32
            
            ' free the 16bit memory block
            Call WOWGlobalUnlockFree16(lpMem16)
          Else
            bError = True
          End If ' lpMem16
        Else
          bError = True
        End If   ' lpfn
        ' free the 16bit module
        Call FreeLibrary16(hMod)
      Else
        bError = True
      End If   ' hMod
    Else
      bError = True
    End If   ' hMod - Else
    
    If bError Then
      ' format the error message and add it to the listbox
      nTotalAdded = nTotalAdded + 1
      
      RaiseEvent ReturnAppletInfo(asApplets(nIndex), 0, "NULL", "Unable to get info for this applet.", 0, 0, cplInvalidApplet, bFailedInit, bIs16Bit, 0)
    End If
  Next ' nIndex
  
  EnumerateApplets = nTotalAdded
  
End Function

Private Sub Class_Initialize()
  m_sngOSVersion = GetOSVersion()
End Sub
